#include "fetchManager.hpp"

#include "client.hpp"

#include <gflags/gflags.h>

DEFINE_uint64(max_incoming_buffered_size, 1024 * 1024 * 200,
    "maximum number of buffered incoming bytes (pending write to storage) per "
    "link");

namespace rtransfer {

namespace {
struct Header {
    std::uint64_t reqId;
    std::uint8_t priority;
    std::uint64_t offset;

    Header(std::unique_ptr<folly::IOBuf> &buf)
    {
        buf->gather(21);
        std::memcpy(&reqId, buf->data(), sizeof(reqId));
        buf->trimStart(sizeof(reqId));
        std::memcpy(&priority, buf->data(), sizeof(priority));
        buf->trimStart(sizeof(priority));
        std::memcpy(&offset, buf->data(), sizeof(offset));
        buf->trimStart(sizeof(offset) + 4);

        reqId = folly::Endian::big(reqId);
        offset = folly::Endian::big(offset);
    }
};
}  // namespace

FetchManager::FetchManager(StoragesMap &storages, HandleCache &handleCache)
    : storages_{storages}
    , handleCache_{handleCache}
{
}

folly::Future<std::size_t> FetchManager::newFetch(std::uint64_t requestId,
    folly::fbstring destStorageId, folly::fbstring destFileId,
    folly::fbstring fileGuid, std::uint64_t offset, std::size_t size,
    folly::Function<void(std::uint64_t, std::size_t)> notifyCb)
{
    VLOG(1) << "New fetch: " << requestId << " " << offset << " " << size;

    auto fi = std::make_shared<FetchInfo>();
    fi->storageId = std::move(destStorageId);
    fi->fileId = std::move(destFileId);
    fi->fileGuid = std::move(fileGuid);
    fi->offset = offset;
    fi->size = size;
    fi->notifyCb = std::move(notifyCb);

    auto result = fetchInfo_.emplace(requestId, std::move(fi));
    DCHECK(result.second);
    ONE_METRIC_COUNTER_INC("fetches.concurrent");
    return result.first->second->promise.getFuture();
}

void FetchManager::read(Context *ctx, std::unique_ptr<folly::IOBuf> buf)
{
    folly::getCPUExecutor()->add(
        [this, buf = std::move(buf)]() mutable { doRead(std::move(buf)); });
}

void FetchManager::doRead(std::unique_ptr<folly::IOBuf> buf)
{
    Header header{buf};
    VLOG(2) << "Received data about " << header.reqId;

    auto it = fetchInfo_.find(header.reqId);
    if (it == fetchInfo_.cend()) {
        LOG(INFO) << "Discarding canceled data for reqId = " << header.reqId;
        return;
    }

    auto &req = *it->second;

    auto storageIt = storages_.find(req.storageId);
    if (storageIt == storages_.cend())
        return cancelFetch(
            it, "no such destination storage " + folly::hexlify(req.storageId));

    folly::IOBufQueue queue{folly::IOBufQueue::cacheChainLength()};
    queue.append(std::move(buf));
    std::size_t toWrite = queue.chainLength();

    handleCache_
        .write(*storageIt->second, req.fileId, req.fileGuid, header.offset,
            std::move(queue), header.priority)
        .then([this, offset = header.offset, toWrite, reqId = header.reqId] {
            auto it = fetchInfo_.find(reqId);
            if (it == fetchInfo_.cend())
                return;

            {
                folly::SpinLockGuard guard{it->second->lock};
                it->second->notifyCb(offset, toWrite);
                it->second->wrote += toWrite;
            }
            maybeFulfillPromise(it);
        })
        .onError([this, reqId = header.reqId](const std::exception &e) {
            auto it = fetchInfo_.find(reqId);
            if (it != fetchInfo_.cend())
                cancelFetch(it, e.what());
        });
}

void FetchManager::cancelFetch(std::uint64_t requestId)
{
    auto it = fetchInfo_.find(requestId);
    if (it == fetchInfo_.cend()) {
        LOG(WARNING) << "Unable to cancel unknown fetch request " << requestId;
        return;
    }

    cancelFetch(it, "canceled");
}

void FetchManager::setTotalSize(std::uint64_t reqId, std::size_t size)
{
    auto it = fetchInfo_.find(reqId);
    if (it == fetchInfo_.cend())
        return;

    {
        folly::SpinLockGuard guard{it->second->lock};
        if (it->second->size < size) {
            LOG(WARNING) << "Refusing to set size to larger than requested";
            return;
        }
        it->second->size = size;
    }
    maybeFulfillPromise(it);
}

void FetchManager::cancelFetch(
    Map::const_iterator it, folly::StringPiece reason)
{
    ONE_METRIC_COUNTER_DEC("fetches.concurrent");
    ONE_METRIC_COUNTER_INC("fetches.cancelled");
    try {
        {
            folly::SpinLockGuard guard{it->second->lock};
            it->second->promise.setException(
                folly::make_exception_wrapper<std::runtime_error>(
                    reason.str()));
        }
        fetchInfo_.erase(it);
    }
    catch (const folly::PromiseAlreadySatisfied &) {
    }
}

void FetchManager::maybeFulfillPromise(Map::const_iterator it)
{
    ONE_METRIC_COUNTER_DEC("fetches.concurrent");

    try {
        bool fulfilled = false;
        {
            folly::SpinLockGuard guard{it->second->lock};
            if (it->second->wrote >= it->second->size) {
                it->second->promise.setValue(it->second->wrote);
                fulfilled = true;
            }
        }
        if (fulfilled)
            fetchInfo_.erase(it);
    }
    catch (const folly::PromiseAlreadySatisfied &) {
    }
}

}  // namespace rtransfer
