#include "reader.hpp"

#include "shaperMap.hpp"

#include <gflags/gflags.h>

DEFINE_uint64(socket_write_codel_interval, 1000, "");
DEFINE_uint64(socket_write_codel_target, 1000, "");

namespace rtransfer {

Reader::Reader(StoragesMap &storages, HandleCache &handleCache,
    std::shared_ptr<ServerSideLink> serverLink)
    : storages_{storages}
    , handleCache_{handleCache}
    , serverLink_{std::move(serverLink)}
    , transportCodel_{
          std::chrono::milliseconds{FLAGS_socket_write_codel_interval},
          std::chrono::milliseconds{FLAGS_socket_write_codel_target}}
{
}

folly::Future<std::size_t> Reader::read(std::uint64_t reqId,
    folly::fbstring storageId, folly::fbstring fileId, folly::fbstring fileGuid,
    std::uint64_t offset, std::size_t size, std::uint8_t priority)
{
    auto it = storages_.find(storageId);
    if (it == storages_.cend())
        return folly::makeFuture<std::size_t>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    VLOG(2) << "Reading " << reqId << " offset: " << offset
            << ", size: " << size;
    return handleCache_
        .read(*it->second, reqId, fileId, fileGuid, offset, size, priority)
        .via(folly::getCPUExecutor().get())
        .then([this, reqId, priority, offset](folly::IOBufQueue buf) {
            std::size_t size = buf.chainLength();
            auto msg =
                constructDataMsg(reqId, offset, priority, std::move(buf));

            auto metric =
                "link." + serverLink_->getPeerAddr() + ".pre_write_buffer.size";
            ONE_METRIC_COUNTER_ADD(metric, size);

            auto conn = serverLink_->chooseDataConnection(size);
            VLOG(2) << "Sending " << reqId << ", offset: " << offset
                    << ", size: " << size << " through conn: " << conn.get();

            auto start = std::chrono::steady_clock::now();
            return conn->write(std::move(msg))
                .via(folly::getCPUExecutor().get())
                .then([=, metric = std::move(metric)] {
                    auto now = std::chrono::steady_clock::now();
                    auto delay = now - start;

                    serverLink_->decreaseSize(conn, size);

                    if (transportOverloaded(now, delay) && shaperMap_)
                        shaperMap_->transportCongestion(now);

                    ONE_METRIC_COUNTER_SUB(metric, size);
                    VLOG(2) << "Sent " << reqId << ", offset: " << offset
                            << ", size: " << size << ", took: "
                            << std::chrono::duration_cast<
                                   std::chrono::milliseconds>(delay)
                                   .count()
                            << "ms";

                    return size;
                });
        });
}

folly::Future<folly::Unit> Reader::cancel(
    std::uint64_t reqId, folly::fbstring storageId)
{
    auto it = storages_.find(storageId);
    if (it == storages_.cend())
        return folly::makeFuture<folly::Unit>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    return it->second->cancel(reqId);
}

template <typename Duration>
bool Reader::transportOverloaded(
    std::chrono::steady_clock::time_point now, Duration duration)
{
    auto delay = std::chrono::duration_cast<std::chrono::nanoseconds>(duration);
    folly::SpinLockGuard guard{codelLock_};
    return transportCodel_.overloaded(now, delay);
}

std::unique_ptr<folly::IOBuf> Reader::constructDataMsg(std::uint64_t reqId,
    std::uint64_t offset, std::uint8_t priority, folly::IOBufQueue buf)
{
    std::uint32_t size = folly::Endian::big<std::uint32_t>(buf.chainLength());
    reqId = folly::Endian::big<std::uint64_t>(reqId);
    offset = folly::Endian::big<std::uint64_t>(offset);

    auto msg = folly::IOBuf::createCombined(
        sizeof(reqId) + sizeof(priority) + sizeof(offset) + sizeof(size));

    auto append = [&](auto val) {
        std::memcpy(msg->writableTail(), &val, sizeof(val));
        msg->append(sizeof(val));
    };

    append(reqId);
    append(priority);
    append(offset);
    append(size);

    if (buf.chainLength() > 0)
        msg->prependChain(buf.move());

    return msg;
}

}  // namespace rtransfer
