#include "serverSideLink.hpp"

#include "shaperMap.hpp"

namespace rtransfer {

ServerSideLink::ServerSideLink(std::string peerAddr)
    : peerAddr_{std::move(peerAddr)}
{
}

std::string ServerSideLink::getPeerAddr() { return peerAddr_; }

void ServerSideLink::attachPipeline(Context *ctx) { ctrlCtx_ = ctx; }

void ServerSideLink::addDataConnection(wangle::DefaultPipeline::Ptr pipeline)
{
    folly::SpinLockGuard guard{lock_};
    dataConns_.emplace_back(Conn{0, pipeline});
}

void ServerSideLink::removeDataConnection(wangle::DefaultPipeline::Ptr pipeline)
{
    LOG(INFO) << "DATA conn down";
    folly::SpinLockGuard guard{lock_};
    auto it = std::find_if(dataConns_.begin(), dataConns_.end(),
        [&](const Conn &c) { return c.second == pipeline; });
    if (it != dataConns_.end()) {
        std::swap(*it, *dataConns_.rbegin());
        dataConns_.pop_back();
    }
}

wangle::DefaultPipeline::Ptr ServerSideLink::chooseDataConnection(
    std::size_t size)
{
    std::size_t oldSize;
    wangle::DefaultPipeline::Ptr conn;
    {
        folly::SpinLockGuard guard{lock_};
        CHECK(!dataConns_.empty());
        auto it = std::min_element(dataConns_.begin(), dataConns_.end());
        std::tie(oldSize, conn) = *it;
        it->first += size;
    }
    VLOG(2) << "Chosen conn " << conn.get() << " (total size: " << size << ")";
    return conn;
}

void ServerSideLink::decreaseSize(
    wangle::DefaultPipeline::Ptr pipeline, std::size_t size)
{
    folly::SpinLockGuard guard{lock_};
    auto it = std::find_if(dataConns_.begin(), dataConns_.end(),
        [&](const Conn &c) { return c.second == pipeline; });
    if (it != dataConns_.end()) {
        DCHECK(it->first >= size);
        it->first -= size;
    }
}

void ServerSideLink::readEOF(Context *ctx)
{
    folly::SpinLockGuard guard{lock_};
    for (auto conn : dataConns_)
        conn.second->getTransport()->getEventBase()->runInEventBaseThread(
            [conn = conn.second] { conn->getTransport()->close(); });
    ctx->fireReadEOF();
}

void ServerSideLink::setShaperMap(ShaperMap *shaperMap)
{
    shaperMap_ = shaperMap;
}

void ServerSideLink::sourceCongestion(const folly::fbstring &storageId)
{
    if (UNLIKELY(!shaperMap_))
        return;

    shaperMap_->sourceCongestion(storageId);
}

DataConn::DataConn(std::shared_ptr<ServerSideLink> link)
    : link_{link}
{
}

void DataConn::attachPipeline(Context *ctx)
{
    link_->addDataConnection(getPipeline(ctx));
}

void DataConn::readEOF(Context *ctx)
{
    link_->removeDataConnection(getPipeline(ctx));
    ctx->fireReadEOF();
}

wangle::DefaultPipeline::Ptr DataConn::getPipeline(Context *ctx)
{
    return std::static_pointer_cast<wangle::DefaultPipeline>(
        ctx->getPipelineShared());
}

ServerSideLinkFactory::ServerSideLinkFactory(
    folly::ConcurrentHashMap<folly::fbstring, std::weak_ptr<ServerSideLink>>
        &map)
    : map_{map}
{
}

std::shared_ptr<ServerSideLink> ServerSideLinkFactory::operator()(
    folly::StringPiece ctrlId, std::string peerAddr)
{
    auto link = std::make_shared<ServerSideLink>(std::move(peerAddr));
    map_.insert_or_assign(ctrlId.fbstr(), link);
    return link;
}

std::shared_ptr<DataConn> ServerSideLinkFactory::createDataConn(
    const folly::fbstring &ctrlId)
{
    auto it = map_.find(ctrlId);
    if (it == map_.cend())
        return {};

    if (auto link = it->second.lock())
        return std::make_shared<DataConn>(link);

    map_.erase(ctrlId);
    return {};
}

}  // namespace rtransfer
