#include "shaperMap.hpp"

#include <functional>

DEFINE_uint64(
    shaper_tick_interval, 100, "interval in milliseconds between ticks");

namespace rtransfer {

ShaperMap::ShaperMap(Reader &reader)
    : reader_{reader}
    , periodicHandler_{std::chrono::milliseconds{FLAGS_shaper_tick_interval},
          std::bind(&ShaperMap::trigger, this)}
{
}

folly::Future<std::size_t> ShaperMap::fetch(MsgPtr msg)
{
    return via(&executor_).then([this, msg = std::move(msg)]() mutable {
        return doFetch(std::move(msg));
    });
}

folly::Future<folly::Unit> ShaperMap::cancel(MsgPtr msg)
{
    return via(&executor_).then([this, msg = std::move(msg)]() mutable {
        return doCancel(std::move(msg));
    });
}

void ShaperMap::congestion(MsgPtr msg)
{
    executor_.add([this, msg = std::move(msg)]() mutable {
        return doCongestion(std::move(msg));
    });
}

void ShaperMap::sourceCongestion(folly::fbstring storageId)
{
    executor_.add([this, storageId = std::move(storageId)]() mutable {
        return doSourceCongestion(std::move(storageId));
    });
}

void ShaperMap::transportCongestion(std::chrono::steady_clock::time_point when)
{
    executor_.add([=] {
        for (auto &shaper : shapers_)
            shaper->signalCongestion(when);
    });
}

folly::Future<std::size_t> ShaperMap::doFetch(MsgPtr msg)
{
    auto key = boost::make_tuple<folly::fbstring, folly::fbstring>(
        {msg->fetch().src()}, {msg->fetch().dest()});

    auto &index = bmi::get<ByKey>(shapers_);
    auto entryIt = index.find(key);
    if (entryIt == index.end()) {
        auto insertResult = index.emplace(std::make_unique<Shaper>(
            boost::get<0>(key), boost::get<1>(key), reader_));
        entryIt = insertResult.first;
    }

    return (*entryIt)->read(std::move(msg));
}

folly::Future<folly::Unit> ShaperMap::doCancel(MsgPtr msg)
{
    auto key = boost::make_tuple<folly::fbstring, folly::fbstring>(
        {msg->cancel().src()}, {msg->cancel().dest()});

    auto &index = bmi::get<ByKey>(shapers_);
    auto entryIt = index.find(key);
    if (entryIt == index.end())
        return folly::makeFuture();

    return (*entryIt)->cancel(std::move(msg));
}

void ShaperMap::doCongestion(MsgPtr msg)
{
    if (msg->congestion().storage_id_case() == proto::Congestion::kSource)
        doNotifyCongestion<BySrc>(msg->congestion().source());
    else
        doNotifyCongestion<ByDst>(msg->congestion().destination());
}

void ShaperMap::doSourceCongestion(folly::fbstring storageId)
{
    doNotifyCongestion<BySrc>(storageId);
}

template <typename Tag>
void ShaperMap::doNotifyCongestion(folly::fbstring storageId)
{
    auto now = std::chrono::steady_clock::now();
    auto &index = bmi::get<Tag>(shapers_);
    auto range = index.equal_range(storageId);
    for (auto it = range.first; it != range.second; ++it)
        (*it)->signalCongestion(now);
}

void ShaperMap::trigger()
{
    executor_.add([this] {
        auto now = std::chrono::steady_clock::now();
        for (auto &shaper : shapers_)
            shaper->trigger(now);
    });
}

}  // namespace rtransfer
