#include "storage.hpp"

#include "uuid.hpp"

#include <folly/FBString.h>
#include <folly/Function.h>
#include <folly/String.h>
#include <folly/executors/Codel.h>
#include <folly/executors/NamedThreadFactory.h>
#include <folly/executors/ThreadPoolExecutor.h>
#include <folly/io/IOBufQueue.h>
#include <gflags/gflags.h>
#include <helpers/storageHelper.h>
#include <monitoring/monitoring.h>

#include <atomic>
#include <chrono>
#include <queue>
#include <unordered_map>

DEFINE_uint64(
    storage_buckets, 100, "max number of in-progress operations on storage");

namespace {
folly::Future<folly::IOBufQueue> readLoop(one::helpers::FileHandlePtr handle,
    std::uint64_t offset, std::size_t size, folly::IOBufQueue buf)
{
    if (size == 0)
        return folly::makeFuture<folly::IOBufQueue>(std::move(buf));

    return handle->read(offset, size)
        .then([=, buf = std::move(buf)](folly::IOBufQueue nextBuf) mutable {
            if (nextBuf.empty()) {
                size = 0;
            }
            else {
                assert(size >= nextBuf.chainLength());
                size -= nextBuf.chainLength();
                offset += nextBuf.chainLength();
                buf.append(std::move(nextBuf));
            }
            return readLoop(std::move(handle), offset, size, std::move(buf));
        });
}

folly::Future<folly::Unit> writeLoop(one::helpers::FileHandlePtr handle,
    std::uint64_t offset, folly::IOBufQueue buf)
{
    if (buf.empty())
        return folly::makeFuture();

    folly::IOBufQueue cloneBuf{folly::IOBufQueue::cacheChainLength()};
    cloneBuf.append(buf.front()->clone());

    return handle->write(offset, std::move(buf))
        .then([=, cloneBuf = std::move(cloneBuf)](std::size_t wrote) mutable {
            cloneBuf.split(wrote);
            return writeLoop(std::move(handle), offset, std::move(cloneBuf));
        });
}
}  // namespace

namespace rtransfer {

Storage::Storage(folly::fbstring storageId,
    one::helpers::StorageHelperPtr storageHelper,
    folly::Function<void()> notifyReadCongestion,
    folly::Function<void()> notifyWriteCongestion)
    : storageIdHex_{folly::hexlify(storageId)}
    , storageHelper_{std::move(storageHelper)}
    , notifyReadCongestion_{std::move(notifyReadCongestion)}
    , notifyWriteCongestion_{std::move(notifyWriteCongestion)}
    , readCodel_{std::chrono::milliseconds{FLAGS_codel_interval},
          std::chrono::milliseconds{FLAGS_codel_target_delay}}
    , writeCodel_{std::chrono::milliseconds{FLAGS_codel_interval},
          std::chrono::milliseconds{FLAGS_codel_target_delay}}
    , readQueue_{FLAGS_storage_buckets}
    , writeQueue_{FLAGS_storage_buckets}
{
    LOG(INFO) << "Adding storage " << storageIdHex_;
}

folly::Future<one::helpers::FileHandlePtr> Storage::open(
    folly::fbstring fileId, OpenType openType)
{
    VLOG(2) << "Opening file '" << fileId << "'";
    const int openFlag = openType == OpenType::read ? O_RDONLY : O_WRONLY;
    return storageHelper_->open(fileId, openFlag, {});
}

folly::Future<folly::IOBufQueue> Storage::read(
    one::helpers::FileHandlePtr handle, std::uint64_t reqId,
    const std::uint64_t offset, const std::size_t size,
    const std::uint8_t priority)
{
    ReadRequest req{reqId, {}, std::chrono::steady_clock::now(),
        std::move(handle), offset, size, std::make_shared<bool>(false)};
    auto future = req.promise.getFuture();
    executor_.add([=, req = std::move(req)]() mutable {
        cancelReq_.emplace(reqId, req.canceled);
        auto size = req.size;
        VLOG(2) << "Pushing " << reqId << " into readQueue";
        auto res = readQueue_.push(priority, std::move(req));
        if (res.second)
            handleRead(res.first);
        else {
            ONE_METRIC_COUNTER_INC(m(".reads.queue.items"));
            ONE_METRIC_COUNTER_ADD(m(".reads.queue.size"), size);
        }
    });
    return future;
}

std::string Storage::m(const char *name)
{
    return "storage." + storageIdHex_.toStdString() + name;
}

folly::Future<folly::Unit> Storage::write(one::helpers::FileHandlePtr handle,
    const std::uint64_t offset, folly::IOBufQueue buf,
    const std::uint8_t priority)
{
    WriteRequest req{{}, std::chrono::steady_clock::now(), std::move(handle),
        offset, std::move(buf)};
    auto future = req.promise.getFuture();
    executor_.add([=, req = std::move(req)]() mutable {
        auto size = req.buf.chainLength();
        auto res = writeQueue_.push(priority, std::move(req));
        if (res.second)
            handleWrite(res.first);
        else {
            ONE_METRIC_COUNTER_INC(m(".writes.queue.items"));
            ONE_METRIC_COUNTER_ADD(m(".writes.queue.size"), size);
        }
    });
    return future;
}

folly::Future<folly::Unit> Storage::cancel(std::uint64_t reqId)
{
    return via(&executor_).then([=] {
        auto range = cancelReq_.equal_range(reqId);
        for (auto it = range.first; it != range.second; ++it)
            *it->second = true;
        cancelReq_.erase(range.first, range.second);
    });
}

void Storage::logPop(Storage::WriteRequest &req) {}

void Storage::logPop(Storage::ReadRequest &req)
{
    VLOG(2) << "Popped " << req.reqId << " from queue";
}

template <typename Queue, typename NotifyCongestion, typename Cb>
void Storage::usingBucket(Queue &queue, const std::size_t bucket, Codel &codel,
    NotifyCongestion &&notifyCongestion, Cb &&cb)
{
    queue.use(bucket);
    auto req = queue.pop(bucket);
    logPop(req);
    auto now = std::chrono::steady_clock::now();
    auto delay = std::chrono::duration_cast<std::chrono::nanoseconds>(
        now - req.requestTime);

    if (codel.overloaded(now, delay))
        notifyCongestion(now);

    auto promise = std::move(req.promise);
    cb(std::move(req))
        .then([promise = std::move(promise)](
                  folly::Try<decltype(cb(std::declval<decltype(req)>()).get())>
                      res) mutable { promise.setTry(std::move(res)); })
        .via(&executor_)
        .ensure([=, &queue, &codel,
                    notifyCongestion =
                        std::forward<NotifyCongestion>(notifyCongestion),
                    cb = std::forward<Cb>(cb)]() mutable {
            if (!queue.empty(bucket))
                usingBucket(queue, bucket, codel, std::move(notifyCongestion),
                    std::move(cb));
            else
                queue.stopUsing(bucket);
        });
}

void Storage::handleRead(const std::size_t bucket)
{
    usingBucket(readQueue_, bucket, readCodel_,
        [this](std::chrono::steady_clock::time_point now) {
            notifyReadCongestion(now);
        },
        [this](ReadRequest req) mutable {
            ONE_METRIC_COUNTER_DEC(m(".reads.queue.items"));
            ONE_METRIC_COUNTER_SUB(m(".reads.queue.size"), req.size);
            if (*req.canceled)
                return folly::makeFuture<folly::IOBufQueue>(
                    folly::make_exception_wrapper<std::runtime_error>(
                        "Operation canceled by user"));

            ONE_METRIC_COUNTER_INC(m(".reads.concurrent"));
            ONE_METRIC_COUNTER_INC(m(".reads.total"));
            return readLoop(req.handle, req.offset, req.size,
                folly::IOBufQueue{folly::IOBufQueue::cacheChainLength()})
                .via(&executor_)
                .ensure([this, reqId = req.reqId] {
                    ONE_METRIC_COUNTER_DEC(m(".reads.concurrent"));
                    cancelReq_.erase(reqId);
                });
        });
}

void Storage::handleWrite(const std::size_t bucket)
{
    usingBucket(writeQueue_, bucket, writeCodel_,
        [this](std::chrono::steady_clock::time_point now) {
            notifyWriteCongestion(now);
        },
        [this](WriteRequest req) mutable {
            ONE_METRIC_COUNTER_DEC(m(".writes.queue.items"));
            ONE_METRIC_COUNTER_SUB(
                m(".writes.queue.size"), req.buf.chainLength());
            ONE_METRIC_COUNTER_INC(m(".writes.concurrent"));
            ONE_METRIC_COUNTER_INC(m(".writes.total"));
            return writeLoop(req.handle, req.offset, std::move(req.buf))
                .ensure([this] {
                    ONE_METRIC_COUNTER_DEC(m(".writes.concurrent"));
                });
        });
}

void Storage::notifyReadCongestion(std::chrono::steady_clock::time_point now)
{
    auto timeSinceLastNotification = now - lastReadCongestionNotif_;
    if (timeSinceLastNotification > std::chrono::milliseconds(500)) {
        lastReadCongestionNotif_ = now;
        notifyReadCongestion_();
    }
}

void Storage::notifyWriteCongestion(std::chrono::steady_clock::time_point now)
{
    auto timeSinceLastNotification = now - lastWriteCongestionNotif_;
    if (timeSinceLastNotification > std::chrono::milliseconds(500)) {
        lastWriteCongestionNotif_ = now;
        notifyWriteCongestion_();
    }
}

}  // namespace rtransfer
