%%%--------------------------------------------------------------------
%%% @author Tomasz Lichon
%%% @copyright (C) 2017 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%--------------------------------------------------------------------
%%% @doc
%%% Manages replica invalidation, which include starting the invalidation and
%%% tracking invalidation's status.
%%% Such gen_server is created for each replica invalidation.
%%% @end
%%%--------------------------------------------------------------------
-module(invalidation_controller).
-author("Tomasz Lichon").

-behaviour(gen_server).

-include("proto/oneprovider/provider_messages.hrl").
-include("modules/datastore/datastore_models.hrl").
-include_lib("ctool/include/logging.hrl").

%% API
-export([mark_finished/1, mark_failed/2]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
    code_change/3]).

-record(state, {
    transfer_id :: transfer:id(),
    session_id :: session:id(),
    file_guid :: fslogic_worker:file_guid(),
    callback :: transfer:callback(),
    space_id :: od_space:id()
}).

%%%===================================================================
%%% API
%%%===================================================================

%%-------------------------------------------------------------------
%% @doc
%% Stops invalidation_controller process and marks invalidation as completed.
%% @end
%%-------------------------------------------------------------------
-spec mark_finished(pid()) -> ok.
mark_finished(Pid) ->
    gen_server2:cast(Pid, finish_invalidation),
    ok.

%%-------------------------------------------------------------------
%% @doc
%% Stops transfer_controller process and marks transfer as failed.
%% @end
%%-------------------------------------------------------------------
-spec mark_failed(pid(), term()) -> ok.
mark_failed(Pid, Error) ->
    gen_server2:cast(Pid, {failed_invalidation, Error}).


%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Initializes the server.
%% @end
%%--------------------------------------------------------------------
-spec init(Args :: term()) ->
    {ok, State :: #state{}} | {ok, State :: #state{}, timeout() | hibernate} |
    {stop, Reason :: term()} | ignore.
init([SessionId, TransferId, FileGuid, Callback]) ->
    ok = gen_server2:cast(self(), start_invalidation),
    {ok, #state{
        transfer_id = TransferId,
        session_id = SessionId,
        file_guid = FileGuid,
        callback = Callback,
        space_id = fslogic_uuid:guid_to_space_id(FileGuid)
    }}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handles call messages.
%% @end
%%--------------------------------------------------------------------
-spec handle_call(Request :: term(), From :: {pid(), Tag :: term()},
    State :: #state{}) ->
    {reply, Reply :: term(), NewState :: #state{}} |
    {reply, Reply :: term(), NewState :: #state{}, timeout() | hibernate} |
    {noreply, NewState :: #state{}} |
    {noreply, NewState :: #state{}, timeout() | hibernate} |
    {stop, Reason :: term(), Reply :: term(), NewState :: #state{}} |
    {stop, Reason :: term(), NewState :: #state{}}.
handle_call(_Request, _From, State) ->
    ?log_bad_request(_Request),
    {reply, wrong_request, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handles cast messages.
%% @end
%%--------------------------------------------------------------------
-spec handle_cast(Request :: term(), State :: #state{}) ->
    {noreply, NewState :: #state{}} |
    {noreply, NewState :: #state{}, timeout() | hibernate} |
    {stop, Reason :: term(), NewState :: #state{}}.
handle_cast(start_invalidation, State = #state{
    transfer_id = TransferId,
    session_id = SessionId,
    file_guid = FileGuid
}) ->
    try
        transfer:mark_active_invalidation(TransferId),
        #provider_response{
            status = #status{code = ?OK}
        } = sync_req:invalidate_file_replica(user_ctx:new(SessionId),
            file_ctx:new_by_guid(FileGuid), undefined, TransferId, undefined
        ),
        {noreply, State}
    catch
        _Error:Reason ->
            ?error_stacktrace("Could not invalidate file ~p due to ~p", [FileGuid, Reason]),
            transfer:mark_failed_invalidation(TransferId),
            {stop, Reason, State}
    end;
handle_cast(finish_invalidation, State = #state{
    transfer_id = TransferId,
    callback = Callback
}) ->
    transfer:mark_completed_invalidation(TransferId),
    notify_callback(Callback),
    {stop, normal, State};
handle_cast({failed_invalidation, Error}, State = #state{
    file_guid = FileGuid,
    transfer_id = TransferId
}) ->
    ?error_stacktrace("Could not invalidate file ~p due to ~p", [FileGuid, Error]),
    transfer:mark_failed_invalidation(TransferId),
    {stop, Error, State};
handle_cast(_Request, State) ->
    ?log_bad_request(_Request),
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handles all non call/cast messages.
%% @end
%%--------------------------------------------------------------------
-spec handle_info(Info :: timeout() | term(), State :: #state{}) ->
    {noreply, NewState :: #state{}} |
    {noreply, NewState :: #state{}, timeout() | hibernate} |
    {stop, Reason :: term(), NewState :: #state{}}.
handle_info(_Info, State) ->
    ?log_bad_request(_Info),
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any
%% necessary cleaning up. When it returns, the gen_server terminates
%% with Reason. The return value is ignored.
%% @end
%%--------------------------------------------------------------------
-spec terminate(Reason :: (normal | shutdown | {shutdown, term()} | term()),
    State :: #state{}) -> term().
terminate(_Reason, #state{
    session_id = SessionId,
    transfer_id = TransferId
}) ->
    session:remove_transfer(SessionId, TransferId).

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Converts process state when code is changed.
%% @end
%%--------------------------------------------------------------------
-spec code_change(OldVsn :: term() | {down, term()}, State :: #state{},
    Extra :: term()) -> {ok, NewState :: #state{}} | {error, Reason :: term()}.
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%===================================================================
%%% Internal functions
%%%===================================================================\


%%--------------------------------------------------------------------
%% @doc
%% Notifies callback about successful transfer
%% @end
%%--------------------------------------------------------------------
-spec notify_callback(transfer:callback()) -> ok.
notify_callback(undefined) -> ok;
notify_callback(<<>>) -> ok;
notify_callback(Callback) ->
    {ok, _, _, _} = http_client:get(Callback).